class
	BINDING_ADDER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			visit_translation_unit,
			end_visit_translation_unit,
			visit_function_declaration,
			end_visit_function_declaration,
			visit_initializer,
			end_visit_function_call,
			end_visit_postfix_expression,
			end_visit_variable_expression,
			end_visit_struct_type,
			end_visit_enum_type,
			post_visit
		end
	BINDINGS_STATUS

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Bindings_active.set_item (True) -- activate binding checking
			Result := True
		end

	description: STRING
		do
			Result := "Setting bindings"
		end

feature {NONE} -- global data

	global_enum_declarations: HASH_TABLE [AST_ENUM_DECLARATION, STRING]
	global_function_declarations: HASH_TABLE [AST_FUNCTION_DECLARATION, STRING]
	global_struct_declarations: HASH_TABLE [AST_STRUCT_DECLARATION, STRING]
	global_variable_declarations: HASH_TABLE [AST_VARIABLE_DECLARATION, STRING]

	visit_translation_unit (node: AST_TRANSLATION_UNIT): BOOLEAN
		do
			create global_enum_declarations.make (16)
			create global_function_declarations.make (16)
			create global_struct_declarations.make (16)
			create global_variable_declarations.make (16)

			across node.declarations as c loop
				if attached {AST_STRUCT_DECLARATION} c.item as struct_declaration then
					global_struct_declarations.extend (struct_declaration, struct_declaration.identifier)
				elseif attached {AST_FUNCTION_DECLARATION} c.item as function_declaration then
					global_function_declarations.extend (function_declaration, function_declaration.identifier)
				elseif attached {AST_VARIABLE_DECLARATION} c.item as variable_declaration then
					global_variable_declarations.extend (variable_declaration, variable_declaration.identifier)
				elseif attached {AST_ENUM_DECLARATION} c.item as enum_declaration then
					global_enum_declarations.extend (enum_declaration, enum_declaration.identifier)
				else
					log ("Unsupported global type: " + c.item.generator)
				end
			end

			Result := True
		end

	 end_visit_translation_unit (node: AST_TRANSLATION_UNIT)
	 	do
			global_enum_declarations := Void
			global_function_declarations := Void
			global_struct_declarations := Void
			global_variable_declarations := Void
	 	end

feature {NONE} -- local data

	local_variable_declarations: HASH_TABLE [AST_VARIABLE_DECLARATION, STRING]

	visit_function_declaration (node: AST_FUNCTION_DECLARATION): BOOLEAN
		do
			if node.is_regular then
				create local_variable_declarations.make (16)

				across node.parameter_list.parameters as c loop
					if attached {AST_VARIABLE_DECLARATION} c.item as tmp then
						local_variable_declarations.extend (tmp, tmp.identifier)
					else
						log ("Unsupported local: " + c.item.generator)
					end
				end
				across node.locals as c loop
					if attached {AST_VARIABLE_DECLARATION} c.item as tmp then
						local_variable_declarations.extend (tmp, tmp.identifier)
					else
						log ("Unsupported local: " + c.item.generator)
					end
				end
			end

			Result := True
		end

	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION)
		do
			if node.is_regular then
				local_variable_declarations := Void
			end
		end

feature {ANY} -- AST_VISITOR

	visit_initializer (node: AST_INITIALIZER): BOOLEAN
		local
			type: AST_TYPE
		do
			if attached {AST_VARIABLE_DECLARATION} node.parent as tmp then
				node.type_binding := tmp.type
			elseif attached {AST_INITIALIZER} node.parent as parent_initializer then
				type := parent_initializer.type_binding
				if attached {AST_ARRAY_TYPE} type or attached {AST_POINTER_TYPE} type then
					node.type_binding := deref(type)
				elseif attached {AST_STRUCT_TYPE} type as struct_type then
					node.type_binding := get_struct_field_type_by_index (struct_type, parent_initializer.elements.index_of (node, 1))
				else
					log ("Unsupported parent type for AST_INITIALIZER: " + type.generator)
				end
			else
				log ("Unsupported parent for AST_INITIALIZER: " + node.parent.generator)
			end
			Result := True
		end

	end_visit_function_call (node: AST_FUNCTION_CALL)
		local
			fd: AST_FUNCTION_DECLARATION
		do
			if node.identifier_str /= Void then
				if global_function_declarations.has_key (node.identifier_str) then
					fd := global_function_declarations[node.identifier_str]
					node.type_binding := fd.return_type
					node.declaration_binding := fd
				else
					log ("Unsupported AST_FUNCTION_CALL: " + node.c_code)
				end
			else
				if attached {AST_FUNCTION_TYPE} node.identifier_expr.type_binding as tmp then
					node.type_binding := tmp.return_type
				else
					log ("Unsupported AST_FUNCTION_CALL: " + node.c_code + " (expression type: " + node.identifier_expr.type_binding.generator + ")")
				end
			end
		end

	end_visit_postfix_expression (node: AST_POSTFIX_EXPRESSION)
		do
			inspect
				node.type
			when C_op_open_square_bracket then
				node.type_binding := deref (node.main_expression.type_binding)
			when C_op_point then
				if attached {AST_STRUCT_TYPE} node.main_expression.type_binding as struct_type then
					node.declaration_binding := get_struct_field_declaration_by_name (struct_type, node.identifier)
					if attached {AST_VARIABLE_DECLARATION} node.declaration_binding as tmp then
						node.type_binding := tmp.type
					elseif attached {AST_BITFIELD} node.declaration_binding as tmp then
						node.type_binding := tmp.type
					end
				else
					log ("Unsupported AST_POSTFIX_EXPRESSION: " + node.c_code)
				end
			when C_op_arrow then
				if
					attached {AST_POINTER_TYPE} node.main_expression.type_binding as pointer_type and then
					attached {AST_STRUCT_TYPE} pointer_type.type as struct_type
				then
					node.declaration_binding := get_struct_field_declaration_by_name (struct_type, node.identifier)
					if attached {AST_VARIABLE_DECLARATION} node.declaration_binding as tmp then
						node.type_binding := tmp.type
					elseif attached {AST_BITFIELD} node.declaration_binding as tmp then
						node.type_binding := tmp.type
					end
				else
					log ("Unsupported AST_POSTFIX_EXPRESSION: " + node.c_code)
				end
			when C_op_increment, C_op_decrement then
				node.type_binding := node.main_expression.type_binding
			end
		end

	end_visit_variable_expression (node: AST_VARIABLE_EXPRESSION)
		local
			vd: AST_VARIABLE_DECLARATION
			fd: AST_FUNCTION_DECLARATION
		do
			if node.declaration_binding = Void then -- enum field expressions already have a binding
				if local_variable_declarations /= Void and then local_variable_declarations.has_key (node.identifier) then
					vd := local_variable_declarations[node.identifier]
					node.declaration_binding := vd
				elseif global_variable_declarations.has_key (node.identifier) then
					vd := global_variable_declarations[node.identifier]
					node.declaration_binding := vd
				elseif global_function_declarations.has_key (node.identifier) then
					fd := global_function_declarations[node.identifier]
					node.declaration_binding := fd
				else
					log ("No binding for " + node.identifier + "?!")
				end
			end
		end

	end_visit_struct_type (node: AST_STRUCT_TYPE)
		do
			node.declaration_binding := global_struct_declarations[node.identifier]
		end

	end_visit_enum_type (node: AST_ENUM_TYPE)
		do
			node.declaration_binding := global_enum_declarations[node.identifier]
		end

	post_visit (node: AST_NODE)
		do
			-- Special cases:
			if attached {AST_EXPRESSION} node as tmp then
				-- Also go through the type bindings. They might have been cloned from other
				-- nodes before those nodes got their type binding (happens in "links").
				tmp.type_binding.accept (Current)
			end
		end

feature {NONE}

	get_struct_field_declaration_by_name (struct_type: AST_STRUCT_TYPE; field_name: STRING_8): AST_STRUCT_FIELD
		require
			struct_type /= Void
			field_name /= Void
		do
			across
				global_struct_declarations[struct_type.identifier].fields as c
			until
				Result /= Void
			loop
				if attached c.item.identifier as identifier and then identifier.is_equal (field_name) then
					Result := c.item
				end
			end
		ensure
			Result /= Void
		end

	get_struct_field_type_by_index (struct_type: AST_STRUCT_TYPE; field_index: INTEGER_32): AST_TYPE
		require
			struct_type /= Void
			field_index > 0
		local
			struct_field: AST_NODE
		do
			struct_field := global_struct_declarations[struct_type.identifier].fields[field_index]
			if attached {AST_VARIABLE_DECLARATION} struct_field as tmp then
				Result := tmp.type
			elseif attached {AST_BITFIELD} struct_field as tmp then
				Result := tmp.type
			end
		ensure
			Result /= Void
		end

end
