class
	BREAK_CONTINUE_RETURN_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			visit_compound_statement,
			visit_function_declaration,
			visit_if_statement,
			visit_labeled_statement,
			visit_switch_statement,
			visit_while_statement,
			end_visit_function_declaration,
			end_visit_jump_statement,
			end_visit_switch_statement,
			end_visit_while_statement
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			create loop_switch_statements.make
			create loop_statements.make

			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING
		do
			Result := "Removing break, continue and return statements"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	visit_compound_statement (node: AST_COMPOUND_STATEMENT): BOOLEAN
		local
			skip_backup: BOOLEAN
			list: LINKED_LIST [AST_STATEMENT]
			index: INTEGER
		do
			inspect mode
			when 1, 2 then -- loop handling or return handling
				from
					list := node.statements
					list.start
				until
					list.after
				loop
					if do_skip then
						index := list.index

						-- surround remaining statements with an if statement
						skip_backup := True
						do_skip := False
						skip_following (node, index)

						-- restore list cursor index to also process new statements.
						list.go_i_th (index)
					end

					list.item.accept (Current)
					if not list.after then
						list.forth
					end
				end

				if skip_backup then
					do_skip := True
				end
				Result := False
			else -- analysis
				Result := True
			end
		end

	visit_function_declaration (node: AST_FUNCTION_DECLARATION): BOOLEAN
		do
			if not node.is_extern then
				--output_stats (1, node)

				needs_return := False
				needs_break := False
				needs_continue := False
				Result := True
			end
		end

	visit_if_statement (node: AST_IF_STATEMENT): BOOLEAN
		local
			skip_backup: BOOLEAN
		do
			inspect mode
			when 1, 2 then -- loop handling or return handling
				node.condition.accept (Current)
				node.then_statements.accept (Current)

				if node.else_statements /= Void then
					-- else branch needs no skip block for jump statements in the then branch
					skip_backup := do_skip
					do_skip := False
					node.else_statements.accept (Current)
					do_skip := skip_backup or do_skip
				end

				Result := False
			else -- analysis
				Result := True
			end
		end

	visit_labeled_statement (node: AST_LABELED_STATEMENT): BOOLEAN
		do
			if mode = 0 and then node.identifier.starts_with ("__Cont") then
				loop_statements.item.has_continue_label := True
			end
			Result := True
		end

	visit_while_statement (node: AST_WHILE_STATEMENT): BOOLEAN
		do
			inspect mode
			when 0, 2 then -- analysis or return handling
				loop_statements.put (node)
				loop_switch_statements.put (node)

				Result := True
			when 1 then -- loop handling
				-- skip nested loops, they are already processed (inside out)
				Result := False
			end
		end

	visit_switch_statement (node: AST_SWITCH_STATEMENT): BOOLEAN
		do
			loop_switch_statements.put (node)
			Result := True
		end

	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION)
		do
			if node.is_regular then
				handle_function_declaration (node)

				if attached concat (", ", <<[needs_return, "l_return"], [needs_break, "l_break"], [needs_continue, "l_continue"]>>) as vars then
					node.jump_variables := vars
				end

				optimize_skip_blocks (node)
				--output_stats(2, node)
			end
		end

	end_visit_jump_statement (node: AST_JUMP_STATEMENT)
		local
			eiffel_stmt: AST_EXPRESSION_STATEMENT
			assignment_expr: AST_ASSIGNMENT_EXPRESSION
		do
			inspect mode
			when 0 then -- analysis
				inspect node.type
				when C_kw_goto then
					if node.goto_identifier.starts_with ("__Cont") then
						loop_statements.item.has_continue_statements := True
					end
				when C_kw_break then
					if not attached {AST_SWITCH_STATEMENT} loop_switch_statements.item then
						loop_statements.item.has_break_statements := True
					end
				when C_kw_continue then
					loop_statements.item.has_continue_statements := True
				when C_kw_return then
					if not loop_statements.is_empty then
						loop_statements.item.has_return_statements := True
					end
				else
					log ("unexpected jump type")
				end
			when 1 then -- loop handling
				if
					node.type = C_kw_break and then
					attached {AST_SWITCH_STATEMENT} loop_switch_statements.item
				then
					-- break belongs to switch statement -> ignore it.
				elseif node.type = C_kw_break then
					-- replace break
					eiffel_stmt := create_eiffel_statement("l_break := True")
					replace_node (node, eiffel_stmt)
					do_skip := True
					move_cursor_to (eiffel_stmt)
				elseif
					node.type = C_kw_continue or else
					(node.type = C_kw_goto and then node.goto_identifier.starts_with ("__Cont"))
				then
					-- replace continue
					eiffel_stmt := create_eiffel_statement("l_continue := True")
					replace_node (node, eiffel_stmt)
					do_skip := True
					move_cursor_to (eiffel_stmt)
				end
			else -- return handling
				if node.type = C_kw_return then
					-- replace `return expr;' with `Result := expr; js.return'
					-- or `return;' with `js.return'
					if node.return_expression /= Void then
						create assignment_expr.make (create {AST_EIFFEL_EXPRESSION}.make ("Result", node.return_expression.type_binding.ast_twin), node.return_expression.ast_twin)
						create eiffel_stmt.make (assignment_expr)
						insert_adjacent_statement (node, eiffel_stmt, False)
					end
					if not does_execution_stop_after (node) then
						eiffel_stmt := create_eiffel_statement("l_return := True")
						replace_node (node, eiffel_stmt)
						needs_return := True
						do_skip := True
					else
						delete_node (node)
					end

					if eiffel_stmt /= Void then
						move_cursor_to (eiffel_stmt)
					end
				end
			end

		end

	end_visit_switch_statement (node: AST_SWITCH_STATEMENT)
		do
			loop_switch_statements.remove
		end

	end_visit_while_statement (node: AST_WHILE_STATEMENT)
		local
			has_return_stmt: BOOLEAN
		do
			inspect mode
			when 0 then -- analysis
				handle_loop (node)

				has_return_stmt := node.has_return_statements
				loop_statements.remove
				loop_switch_statements.remove

				if not loop_statements.is_empty then
					-- propagate has_return_statement to outer loop
					loop_statements.item.has_return_statements := True
				end
			when 2 then -- return handling
				node.has_return_statements := False
				loop_statements.remove
				loop_switch_statements.remove
			else -- loop handling
			end
		end

feature {NONE} -- Implementation

	mode: INTEGER
		-- 0: analysis
		-- 1: loop (break and continue) handling
		-- 2: return handling

	do_skip: BOOLEAN
		-- skip rest of compound statement

	loop_switch_statements: LINKED_STACK [AST_STATEMENT]
		-- stack of nested loops and switch statements (allows to check if a `break' belongs to switch or loop)

	loop_statements: LINKED_STACK [AST_WHILE_STATEMENT]
		-- stack of nested loops

	needs_return, needs_break, needs_continue: BOOLEAN
		-- does the current function declaration need the variables?

feature {NONE} -- Implementation

	move_cursor_to (node: AST_STATEMENT)
		do
			if attached {AST_COMPOUND_STATEMENT} node.parent as comp then
				if
					comp.statements.after or else
					comp.statements.before or else
					comp.statements.item /= node
				then
					comp.statements.start
					comp.statements.search (node)
				end
			elseif attached {AST_LABELED_STATEMENT} node.parent as label then
				move_cursor_to (label)
			else
				log ("Unexpected parent of AST_STATEMENT")
			end
		end

	negated_eiffel_expression (eiffel_code: STRING): AST_UNARY_EXPRESSION
		require
			eiffel_code /= Void and then not eiffel_code.is_empty
		local
			eiffel_expr: AST_EIFFEL_EXPRESSION
		do
			create eiffel_expr.make (eiffel_code, create {AST_BASIC_TYPE}.make (Ast_bt_bool))
			create Result.make (C_op_not, eiffel_expr)
		end

	create_eiffel_statement (eiffel_code: STRING): AST_EXPRESSION_STATEMENT
		require
			eiffel_code /= Void and then not eiffel_code.is_empty
		local
			eiffel_expr: AST_EIFFEL_EXPRESSION
		do
			create eiffel_expr.make (eiffel_code, create {AST_BASIC_TYPE}.make (Ast_bt_void))
			create Result.make (eiffel_expr)
		end

	add_condition (current_condition: AST_EXPRESSION; additional_condition: AST_EXPRESSION)
			-- add `additional_condition' to `current_condition' or replace `True'
			-- in the first case the resulting condition will be: `additional_condition && (current_condition)'
		local
			new_condition: AST_GENERIC_EXPRESSION
		do
			if
				attached {AST_BOOLEAN_CONSTANT} current_condition as const and then
				const.value = True
			then
				replace_node (current_condition, additional_condition)
			else
				-- add additional condition as first one, the current condition may not be executable!
				-- for example (with p a pointer): `if (p=0) {break}; if (p->i > 0) {(p->i)++}'
				create new_condition.make (C_op_logical_and, additional_condition, create {AST_PARENTHESIZED_EXPRESSION}.make (current_condition.ast_twin))
				replace_node (current_condition, new_condition)
			end
		end

	skip_following (node: AST_COMPOUND_STATEMENT; first: INTEGER)
			-- all statements from index `first' onwards are moved into a new skip block
		require
			not node.statements.after
			node.statements.valid_index (first)
		local
			list: LINKED_LIST [AST_STATEMENT]
			inner_list: ARRAYED_LIST [AST_STATEMENT]
			inside_switch: BOOLEAN
			condition: AST_UNARY_EXPRESSION
		do
			if mode = 2 then
				condition := negated_eiffel_expression ("l_return")
			else
				condition := negated_eiffel_expression ("(" + concat (" or ", <<[needs_return, "l_return"], [needs_break, "l_break"], [needs_continue, "l_continue"]>>) + ")")
			end

			list := node.statements
			inside_switch := loop_switch_statements.count /= 0 and then attached {AST_SWITCH_STATEMENT} loop_switch_statements.item

			-- move statements to inner list
			from
				create inner_list.make (0)
				list.go_i_th (first)
			until
				list.after or else
				(inside_switch and then
				attached {AST_JUMP_STATEMENT} list.item as jump_stmt and then
				jump_stmt.type = C_kw_break)
			loop
				if
					attached {AST_SWITCH_CASE} list.item or
					attached {AST_SWITCH_DEFAULT} list.item
				then
					-- do not include switch labels in skip block:
					-- finish current block and start new one
					add_skip_block (node, condition.ast_twin, inner_list)
					create inner_list.make (0)
					list.forth
				else
					inner_list.extend (list.item.ast_twin)
					list.remove
				end
			end

			add_skip_block (node, condition, inner_list)
		end

	add_skip_block (node: AST_COMPOUND_STATEMENT; condition: AST_EXPRESSION; list: ARRAYED_LIST [AST_STATEMENT])
			-- create a skip block with the statements form `list' and add it to `node.statements' at the current position
		local
			skip_block: AST_IF_STATEMENT
		do
			if
				list.count = 1 and then
				attached {AST_IF_STATEMENT} list.first as stmt and then
				stmt.is_skip_block
			then
				-- keep existing skip block -> reinsert
				stmt.parent := node
				node.statements.put_left (stmt)
			elseif not list.is_empty then
				-- create skip block and add it to the list
				create skip_block.make_if (condition, create {AST_COMPOUND_STATEMENT}.make_with_statements (list.to_array))
				skip_block.is_skip_block := True
				skip_block.parent := node
				node.statements.put_left (skip_block)
			end
		end

	handle_loop (node: AST_WHILE_STATEMENT)
			-- remove break and continue in the loop `node'
		local
			for_loop_increment: ARRAYED_LIST [AST_STATEMENT]
			last_statment: AST_STATEMENT
			l_tmp: AST_EXPRESSION_STATEMENT
		do
			if
				node.statements.statements.count = 1 and then
				attached {AST_BOOLEAN_CONSTANT} node.condition as const and then
				const.value = True and then
				attached {AST_JUMP_STATEMENT} node.statements.statements.first as stmt and then
				stmt.type = C_kw_break
			then
				-- remove `while(1){break;}'
				if attached {AST_LABELED_STATEMENT} node.parent as label then
					replace_node (node, create {AST_EXPRESSION_STATEMENT})
				else
					delete_node (node)
				end
			elseif
				node.has_return_statements or
				node.has_break_statements or
				node.has_continue_statements
			then
				needs_return := needs_return or node.has_return_statements
				needs_break := needs_break or node.has_break_statements
				needs_continue := needs_continue or node.has_continue_statements

				-- add new loop conditional
				if node.has_break_statements or else node.has_return_statements then
					-- add `not js.loop_done' to loop conditional or replace `True'
					add_condition (node.condition, negated_eiffel_expression ("(" + concat (" or ", <<[node.has_return_statements, "l_return"], [node.has_break_statements, "l_break"]>>) + ")"))
				end

				-- reset `l_continue' and `l_break' if necessary
				if node.has_continue_statements then
					create l_tmp.make (create {AST_EIFFEL_EXPRESSION}.make ("l_continue := False", create {AST_BASIC_TYPE}.make (Ast_bt_void)))
					l_tmp.parent := node.statements
					node.statements.statements.put_front (l_tmp)

					insert_adjacent_statement (node, l_tmp.ast_twin, True)
				end
				if node.has_break_statements then
					create l_tmp.make (create {AST_EIFFEL_EXPRESSION}.make ("l_break := False", create {AST_BASIC_TYPE}.make (Ast_bt_void)))
					l_tmp.parent := node.statements
					node.statements.statements.put_front (l_tmp)

					insert_adjacent_statement (node, l_tmp.ast_twin, True)
				end

				-- extract for loop increment statements
				if node.has_continue_label then
					for_loop_increment := extract_for_loop_increment (node)
					node.has_continue_label := False
				end

				-- remove break and continue statements and add necessary skip blocks
				mode := 1
				node.statements.accept (Current)
				mode := 0
				do_skip := False

				-- add for loop increment statements
				if for_loop_increment /= Void then
					if node.has_return_statements or node.has_break_statements then
						-- use `js.loop_increment' as condition
						add_skip_block (node.statements, create {AST_EIFFEL_EXPRESSION}.make (
								"(" + concat (" and ", <<[node.has_return_statements, "not l_return"], [node.has_break_statements, "not l_break"]>>) + ")",
								create {AST_BASIC_TYPE}.make (Ast_bt_bool)
							), for_loop_increment)
					else
						across
						for_loop_increment as c
						loop
							last_statment := node.statements.statements.last
							insert_adjacent_statement (last_statment, c.item, True)
						end
					end
				end

				node.has_break_statements := False
				node.has_continue_statements := False
			else
				-- nothing to change
			end
		end

	extract_for_loop_increment (node: AST_WHILE_STATEMENT): ARRAYED_LIST [AST_STATEMENT]
			-- removes all statements after the continue label and returns them in a list
		require
			node.has_continue_label
		local
			statement_list: LINKED_LIST [AST_STATEMENT]
			increment_list: ARRAYED_LIST [AST_STATEMENT]
			extract: BOOLEAN
		do
			statement_list := node.statements.statements
			create increment_list.make (0)

			from
				statement_list.start
			until
				statement_list.after
			loop
				if extract then
					increment_list.extend (statement_list.item.ast_twin)
					statement_list.remove
				elseif
					attached {AST_LABELED_STATEMENT} statement_list.item as label and then
					label.identifier.starts_with ("__Cont")
				then
					extract := True
					increment_list.extend (label.statement.ast_twin)
					statement_list.remove
				else
					statement_list.forth
				end
			end

			Result := increment_list
		end

	handle_function_declaration (node: AST_FUNCTION_DECLARATION)
			-- transform return
		do
			-- replace return statements and add necessary skip blocks
			mode := 2
			node.statements.accept (Current)
			mode := 0
			do_skip := False

			node.return_statements_removed := True
		end

	output_stats (run: INTEGER; node: AST_FUNCTION_DECLARATION)
		local
			stat_visitor: AST_STATISTICS_VISITOR
		do
			create stat_visitor
			node.accept (stat_visitor)
			log (run.out + ", " + node.identifier + ", " + stat_visitor.to_csv_string)
		end

	does_execution_stop_after (node: AST_STATEMENT): BOOLEAN
		require
			node /= Void
		local
			jump_stmt: AST_JUMP_STATEMENT
		do
			if attached {AST_COMPOUND_STATEMENT} node.parent as compound_statement then
				if attached {AST_SWITCH_STATEMENT} compound_statement.parent as switch_statement then
					if does_execution_stop_after (switch_statement) then
						-- check if it is last statement in switch
						if compound_statement.statements.last = node then
							Result := True
						else
							jump_stmt ?= get_adjacent_statement (node, True)
							if jump_stmt /= Void and then jump_stmt.type = C_kw_break then
								Result := True
							end
						end
					end
				elseif compound_statement.statements.last /= node or attached {AST_WHILE_STATEMENT} compound_statement.parent then
					Result := False
				elseif attached {AST_IF_STATEMENT} compound_statement.parent as if_statement then
					Result := does_execution_stop_after (if_statement)
				elseif attached {AST_FUNCTION_DECLARATION} compound_statement.parent then
					Result := True
				else
					log ("[does_execution_stop_after] Unexpected case?!")
				end
			elseif attached {AST_LABELED_STATEMENT} node.parent as labeled_statement then
				Result := does_execution_stop_after (labeled_statement)
			else
				log (generator + ".does_execution_stop_after: Unexpected parent: " + node.parent.generator + " (" + node.parent.c_code + ")")
			end
		end


feature {NONE} -- Optimization

	same_condition (first, second: AST_EXPRESSION): BOOLEAN
		require
			first /= Void
			second /= Void
		do
			Result := first.eiffel_code.is_equal (second.eiffel_code)
		end

	optimize_skip_blocks (node: AST_FUNCTION_DECLARATION)
		require
			node /= Void
		local
			optimizer: AST_AGENT_VISITOR
		do
			create optimizer
			optimizer.agents.visit_if_statement := agent (ag_node: AST_IF_STATEMENT; ag_visitor: AST_AGENT_VISITOR): BOOLEAN
					-- remove nesting of skip blocks where possible
				local
					moved_skip_block: AST_IF_STATEMENT
				do
					if
						ag_node.is_skip_block and then
						attached {AST_COMPOUND_STATEMENT} ag_node.parent as comp and then
						comp.statements.last = ag_node and then
						attached {AST_IF_STATEMENT} comp.parent as parent_if_stmt and then
						parent_if_stmt.is_skip_block and then
						same_condition (ag_node.condition, parent_if_stmt.condition)
					then
						-- remove nesting of skip blocks
						-- transforms `skip_block {foo; skip_block {bar}}' to `skip_block {foo}; skip_block {bar}'
						moved_skip_block := ag_node.ast_twin
						insert_adjacent_statement (parent_if_stmt, moved_skip_block, True)
						delete_node (ag_node)
						ag_node.parent := Void
						moved_skip_block.accept (ag_visitor)
						Result := False
					else
						Result := True
					end
				end (?, optimizer)

			optimizer.agents.end_visit_if_statement := agent (ag_node: AST_IF_STATEMENT)
					-- combine if statement and skip block where possible
				local
					list: LIST [AST_STATEMENT]
					new_skip_block: AST_IF_STATEMENT
				do
					if
						ag_node.parent /= Void and then
						ag_node.is_skip_block
					then
						list := ag_node.then_statements.statements
						if
							list.count = 1 and then
							attached {AST_IF_STATEMENT} list.first as if_stmt and then
							if_stmt.else_statements = Void
						then
							-- combine if statement with skip block by adding the if condition to the skip block condition
							add_condition (if_stmt.condition, ag_node.condition.ast_twin)
							create new_skip_block.make_if (if_stmt.condition.ast_twin, if_stmt.then_statements.ast_twin)
							new_skip_block.is_skip_block := True
							replace_node (ag_node, new_skip_block)
						end
					end
				end

			node.accept (optimizer)
		end

feature {NONE}

	concat (a_separator: STRING_8; a_strings: ARRAY [TUPLE [use_string: BOOLEAN; string: STRING_8]]): STRING_8
		require
			a_separator /= Void and then not a_separator.is_empty
			a_strings /= Void and then not a_strings.is_empty
			across a_strings as c all c.item.string /= Void end
		local
			l_first: BOOLEAN
		do
			l_first := True
			create Result.make_empty
			across a_strings as c loop
				if c.item.use_string then
					if l_first then
						l_first := False
					else
						Result.append (a_separator)
					end
					Result.append (c.item.string)
				end
			end
			if Result.is_empty then
				Result := Void
			end
		end

invariant
	mode = 0 or mode = 1 or mode = 2

end
